'''Tests replicating example notebooks.'''
# pylint: disable=missing-function-docstring,missing-class-docstring

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from glob import glob
import os
from pathlib import Path
import subprocess

from ngautonml.wrangler.wrangler import Wrangler
from ngautonml.problem_def.problem_def import ProblemDefinition
from ngautonml.algorithms.impl.algorithm_auto import FakeAlgorithmCatalogAuto
# pylint: disable=duplicate-code


def module_path() -> Path:
    return Path(__file__).parents[3]


def flowers_config():
    train_dir = module_path() / 'examples' / 'flowers' / 'train'
    test_dir = module_path() / 'examples' / 'flowers' / 'test'
    pdef = {
        "dataset": {
            "config": "image_tensor",
            "train_dir": train_dir,
            "test_dir": test_dir
        },
        "problem_type": {
            "data_type": "IMAGE",
            "task": "MULTICLASS_CLASSIFICATION"
        },
        "cross_validation": {
            "k": 1
        },
        "hyperparams": [
            {
                "_comments": [
                    "We need fewer epochs for a test environment without GPU."
                ],
                "select": {
                    "algorithm": "tf.keras.sequential"
                },
                "params": {
                    "epochs": {
                        "fixed": 2
                    }
                }
            }
        ]
    }
    return ProblemDefinition(pdef)


def test_wrangle_image() -> None:
    plugin_root = Path(__file__).parents[1]
    env = os.environ.copy()

    subprocess.run(['python', '-m', 'build'], cwd=plugin_root, check=True)

    pluginpaths = glob(str(plugin_root / 'dist' / 'ngautonml_image-*py3-none-any.whl'))
    assert len(pluginpaths) == 1, f'BUG: unexpected extra whl files: {pluginpaths}'
    pluginpath = Path(pluginpaths[0])
    subprocess.run(['pip', 'install', pluginpath], check=True, env=env)

    problem_definition = flowers_config()
    dut = Wrangler(
        problem_definition=problem_definition,
        algorithm_catalog=FakeAlgorithmCatalogAuto
    )
    got = dut.fit_predict_rank()

    assert list(got.train_results.values())[0].prediction.predictions.shape == (733, 1)

    subprocess.run(['pip', 'uninstall', '-y', pluginpath], check=False)



'''Params parser for ImageTensorDataLoader.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
from typing import Set

from ngautonml.config_components.impl.config_component import ConfigComponent


class ImageTensorParams(ConfigComponent):
    '''Params parser for ImageTensorDataLoader'''

    class Keys(Enum):
        '''Child classes need to define this class with all their keys.'''
        TRAIN_DIR = 'train_dir'
        TEST_DIR = 'test_dir'
        VALIDATION_SPLIT = 'validation_split'
        SEED = 'seed'
        IMG_HEIGHT = 'img_height'
        IMG_WIDTH = 'img_width'
        BATCH_SIZE = 'batch_size'

    Defaults = {
        Keys.TRAIN_DIR: None,
        Keys.TEST_DIR: None,
        Keys.VALIDATION_SPLIT: 0.2,
        Keys.SEED: 123,
        Keys.IMG_HEIGHT: 128,
        Keys.IMG_WIDTH: 128,
        Keys.BATCH_SIZE: 32,
    }

    def required_keys(self) -> Set[str]:
        keys = super().required_keys()
        keys.update({
            self.Keys.TRAIN_DIR.value
        })
        return keys

    def validate(self, **kwargs) -> None:
        pass

    @property
    def train_dir(self):
        '''train_dir'''
        return self._get_with_default(self.Keys.TRAIN_DIR.value,
                                      dflt=self.Defaults[self.Keys.TRAIN_DIR])

    @property
    def test_dir(self):
        '''test_dir'''
        return self._get_with_default(self.Keys.TEST_DIR.value,
                                      dflt=self.Defaults[self.Keys.TEST_DIR])

    @property
    def validation_split(self):
        '''validation_split'''
        return self._get_with_default(self.Keys.VALIDATION_SPLIT.value,
                                      dflt=self.Defaults[self.Keys.VALIDATION_SPLIT])

    @property
    def seed(self):
        '''seed'''
        return self._get_with_default(self.Keys.SEED.value,
                                      dflt=self.Defaults[self.Keys.SEED])

    @property
    def img_height(self):
        '''img_height'''
        return self._get_with_default(self.Keys.IMG_HEIGHT.value,
                                      dflt=self.Defaults[self.Keys.IMG_HEIGHT])

    @property
    def img_width(self):
        '''img_width'''
        return self._get_with_default(self.Keys.IMG_WIDTH.value,
                                      dflt=self.Defaults[self.Keys.IMG_WIDTH])

    @property
    def batch_size(self):
        '''batch_size'''
        return self._get_with_default(self.Keys.BATCH_SIZE.value,
                                      dflt=self.Defaults[self.Keys.BATCH_SIZE])



'''Datasetloader for image classification, using tensorflow format.

Images are stored in subdirectories whose names indicate thier class.
'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Optional

import tensorflow as tf


from ngautonml.config_components.dataset_config import DatasetConfig
from ngautonml.config_components.impl.config_component import ValidationErrors
from ngautonml.data_loaders.impl.data_loader import DataLoader
from ngautonml.data_loaders.impl.data_loader_catalog import DataLoaderCatalog
from ngautonml.problem_def.output_config import ConfigError
from ngautonml.wrangler.constants import Defaults
from ngautonml.wrangler.dataset import Column, Dataset, Metadata, RoleName

from .impl.image_tensor_params import ImageTensorParams


class ImageTensorDataLoader(DataLoader):
    '''Holds information about an image classification dataset.

    Images are stored in subdirectories whose names indicate thier class.
    '''
    name = 'image_tensor'
    tags: Dict[str, List[str]] = {
        'input_format': ['image_directory'],
        'loaded_format': ['tensorflow_tensor']
    }
    _config: DatasetConfig
    _metadata: Metadata
    _params: ImageTensorParams

    def __init__(self, config: DatasetConfig):
        super().__init__(config=config)
        self._params = ImageTensorParams(self._config.params)
        # If no target is specified, set it to the default target for image classification.
        roles = self._metadata.roles
        if RoleName.TARGET not in roles or len(roles[RoleName.TARGET]) == 0:
            roles = self._metadata.roles
            # TODO(Merritt): plugins should have thier own defaults & constants
            roles[RoleName.TARGET] = [Column(name=Defaults.IMAGE_CLF_TARGET_NAME)]
            self._metadata = self._metadata.override_roles(roles=roles)

    def validate(self, dataset: Optional[Dataset]) -> None:
        errors: List[ConfigError] = []

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)

    def _load_train(self) -> Dataset:
        train_ds = tf.keras.utils.image_dataset_from_directory(
            self._params.train_dir,
            validation_split=float(self._params.validation_split),
            subset="training",
            seed=self._params.seed,
            image_size=(int(self._params.img_height),
                        int(self._params.img_width)),
            batch_size=int(self._params.batch_size)
        )

        val_ds = tf.keras.utils.image_dataset_from_directory(
            self._params.train_dir,
            validation_split=float(self._params.validation_split),
            subset="validation",
            seed=self._params.seed,
            image_size=(int(self._params.img_height),
                        int(self._params.img_width)),
            batch_size=int(self._params.batch_size)
        )
        dataset = Dataset(
            metadata=self._metadata,
            keras_ds=train_ds,
            keras_validate=val_ds
        )

        return dataset

    def _load_test(self) -> Optional[Dataset]:
        test_ds = tf.keras.utils.image_dataset_from_directory(
            self._params.test_dir,
            labels=None,
            seed=self._params.seed,
            image_size=(int(self._params.img_height),
                        int(self._params.img_width)),
            batch_size=int(self._params.batch_size),
            shuffle=False
        )

        dataset = Dataset(
            metadata=self._metadata,
            keras_ds=test_ds
        )
        return dataset

    def _dataset(self, data: Any, **kwargs) -> Dataset:
        # TODO(Merritt): implement this
        raise NotImplementedError

    def poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        return self.load_train()

    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        # TODO(Piggy): implement this
        raise NotImplementedError


def register(catalog: DataLoaderCatalog, *unused_args, **unused_kwargs):
    '''Register all the objects in this file.'''
    catalog.register(ImageTensorDataLoader)



'''Tests for image_dir.py.'''

import math
from pathlib import Path

# pylint: disable=no-name-in-module
from tensorflow.python.data.ops.dataset_ops import BatchDataset  # type: ignore[import]

from ngautonml.config_components.dataset_config import DatasetConfig

from .image_tensor_data_loader import ImageTensorDataLoader


NUM_IMAGES = 3665

MINIMAL_CLAUSE = {
    'config': 'image_dir',
    'params': {
        'train_dir': str(Path(__file__).parents[4] / 'examples' / 'flowers' / 'train')
    }
}


def test_defaults() -> None:
    '''Test that default parameters get applied when not specified.'''
    config = DatasetConfig(MINIMAL_CLAUSE)
    dut = ImageTensorDataLoader(config)
    dataset = dut.load_train()
    assert dataset is not None
    got = dataset['keras_ds']
    assert isinstance(got, BatchDataset)
    assert got.class_names == ['daisy', 'dandelion', 'roses', 'sunflowers', 'tulips']
    for image_batch, _ in got:
        # shape tuple is (batch_size, img_height, img_width, _)
        assert image_batch.shape == (32, 128, 128, 3)
        break
    want_val_split = 1 - 0.2
    num_images_got = len([1 for _ in got.unbatch()])
    num_images_want = math.floor(NUM_IMAGES * want_val_split)
    assert num_images_got == num_images_want


OVERRIDE_CLAUSE = {
    'config': 'image_dir',
    'params': {
        'train_dir': str(Path(__file__).parents[4] / 'examples' / 'flowers' / 'train'),
        'validation_split': 0.6,
        'img_height': 2,
        'img_width': 8,
        'batch_size': 10
    }
}


def test_overrides() -> None:
    '''Test that overridden parameters get properly applied.'''
    config = DatasetConfig(OVERRIDE_CLAUSE)
    dut = ImageTensorDataLoader(config=config)
    dataset = dut.load_train()
    assert dataset is not None
    got = dataset['keras_ds']
    assert isinstance(got, BatchDataset)
    for image_batch, _ in got:
        # shape tuple is (batch_size, img_height, img_width, _)
        assert image_batch.shape == (10, 2, 8, 3)
        break
    want_val_split = 1 - 0.6
    num_images_got = len([1 for _ in got.unbatch()])
    num_images_want = math.floor(NUM_IMAGES * want_val_split)
    assert num_images_got == num_images_want



'''Tests for Sequential algorithm for TensorFlow Keras.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Tuple

import pandas as pd
import pytest
import tensorflow as tf

from ngautonml.metrics.impl.metric_auto import MetricCatalogAuto
from ngautonml.problem_def.task import TaskType, DataType
from ngautonml.wrangler.dataset import Column, Dataset, DatasetKeys, Metadata, RoleName

from .sequential import KerasSequential
# pylint: disable=missing-function-docstring, protected-access


def load_classification_dataset() -> Tuple[Dataset, Dataset, Dataset]:
    batch_size = 32
    img_height = 180
    img_width = 180

    train_dir = 'examples/flowers/train/'
    test_dir = 'examples/flowers/test'

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.MULTICLASS_CLASSIFICATION,
        data_type=DataType.IMAGE
    )

    train_ds = tf.keras.utils.image_dataset_from_directory(
        train_dir,
        validation_split=0.2,
        subset="training",
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size)
    assert isinstance(train_ds, tf.data.Dataset)

    val_ds = tf.keras.utils.image_dataset_from_directory(
        train_dir,
        validation_split=0.2,
        subset="validation",
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size)
    assert isinstance(val_ds, tf.data.Dataset)

    test_ds = tf.keras.utils.image_dataset_from_directory(
        test_dir,
        labels=None,
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size,
        shuffle=False)

    assert isinstance(test_ds, tf.data.Dataset)

    dataset = Dataset(
        metadata=metadata,
        keras_ds=train_ds,
        keras_validate=val_ds
    )

    test = Dataset(
        metadata=metadata,
        keras_ds=test_ds
    )

    truthiness = pd.DataFrame(['daisy', 'dandelion', 'roses', 'sunflowers', 'tulips'],
                              columns=['target'])
    truth = Dataset(
        metadata=metadata,
        ground_truth=truthiness
    )
    return dataset, test, truth


def test_sunny_day() -> None:

    tf.keras.utils.set_random_seed(123)
    tf.config.experimental.enable_op_determinism()
    metric_catalog = MetricCatalogAuto()

    trn_data, test_data, ground_truth = load_classification_dataset()

    class_names = trn_data[DatasetKeys.KERAS_DS.value].class_names

    num_classes = len(class_names)

    algorithm = KerasSequential(img_height=180, img_width=180, epochs=2, num_classes=num_classes)
    dut = algorithm.instantiate()

    dut.fit(trn_data)

    result = dut.predict(test_data)

    metric = metric_catalog.lookup_by_name('accuracy_score')
    assert metric.calculate(pred=result,
                            ground_truth=ground_truth
                            ) == pytest.approx(0.8, 0.26)
